/*
 * Copyright (c) 2017 Ted Unangst <tedu@tedunangst.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
package main

import (
	"log"
	"reflect"
	"sync"

	"github.com/yuin/gopher-lua"
)

type Interpreter struct {
	L *lua.LState
}

var runqmtx sync.Mutex
var runnerqueue []*Interpreter

func getinterpreter() *Interpreter {
	runqmtx.Lock()
	var runner *Interpreter
	if len(runnerqueue) > 0 {
		defer runqmtx.Unlock()
		runner = runnerqueue[len(runnerqueue)-1]
		runnerqueue = runnerqueue[0 : len(runnerqueue)-1]
	} else {
		runqmtx.Unlock()
		L := lua.NewState()
		addfunctions(L)
		err := L.DoFile("filter.lua")
		if err != nil {
			log.Printf("error running file %s\n", err)
		}
		runner = &Interpreter{
			L: L,
		}
	}
	return runner
}

func putinterpreter(runner *Interpreter) {
	alwaysreload := true
	if alwaysreload {
		runner.L.Close()
	} else {
		runqmtx.Lock()
		defer runqmtx.Unlock()
		runnerqueue = append(runnerqueue, runner)
	}
}

func func2lua(gofunc interface{}) lua.LGFunction {
	return func(L *lua.LState) int {
		rf := reflect.ValueOf(gofunc)
		ft := reflect.TypeOf(gofunc)
		numargs := ft.NumIn()
		var args []reflect.Value
		for i := 0; i < numargs; i++ {
			var gv interface{}
			lt := L.Get(i + 1).Type()
			switch lt {
			case lua.LTString:
				gv = L.ToString(i + 1)
			case lua.LTNumber:
				gv = L.ToNumber(i + 1)
			case lua.LTBool:
				gv = L.ToBool(i + 1)
			case lua.LTUserData:
				gv = L.ToUserData(i + 1).Value
			case lua.LTNil:
				gv = reflect.Zero(ft.In(i)).Interface()
			default:
				panic("can't convert type from lua")
			}
			args = append(args, reflect.ValueOf(gv))
		}
		rvs := rf.Call(args)
		for i := 0; i < len(rvs); i++ {
			pushsomething(L, rvs[i])
		}
		return len(rvs)
	}
}

func (runner *Interpreter) Call(funcname string, args ...interface{}) string {
	L := runner.L
	L.Push(L.GetGlobal(funcname))
	for i := 0; i < len(args); i++ {
		pushsomething(L, reflect.ValueOf(args[i]))
	}
	L.Call(len(args), 1)
	rv := L.ToString(1)
	L.Pop(1)
	return rv
}

func pushsomething(L *lua.LState, rv reflect.Value) {
	switch rv.Kind() {
	case reflect.Bool:
		L.Push(lua.LBool(rv.Bool()))
	case reflect.Int:
	case reflect.Int8:
	case reflect.Int16:
	case reflect.Int32:
	case reflect.Int64:
	case reflect.Uint:
	case reflect.Uint8:
	case reflect.Uint16:
	case reflect.Uint32:
	case reflect.Uint64:
		L.Push(lua.LNumber(rv.Int()))
	case reflect.Float32:
	case reflect.Float64:
		L.Push(lua.LNumber(rv.Float()))
	case reflect.String:
		L.Push(lua.LString(rv.String()))
	case reflect.Slice:
		table := L.NewTable()
		for i := 0; i < rv.Len(); i++ {
			appendsomething(L, table, rv.Index(i))
		}
		L.Push(table)
	case reflect.Ptr:
		if rv.IsNil() {
			L.Push(lua.LNil)
		} else {
			ud := L.NewUserData()
			ud.Value = rv.Interface()
			L.Push(ud)
		}
	default:
		ud := L.NewUserData()
		ud.Value = rv.Interface()
		L.Push(ud)
	}
}

func appendsomething(L *lua.LState, table *lua.LTable, rv reflect.Value) {
	switch rv.Kind() {
	case reflect.Bool:
		table.Append(lua.LBool(rv.Bool()))
	case reflect.Int:
	case reflect.Int8:
	case reflect.Int16:
	case reflect.Int32:
	case reflect.Int64:
	case reflect.Uint:
	case reflect.Uint8:
	case reflect.Uint16:
	case reflect.Uint32:
	case reflect.Uint64:
		table.Append(lua.LNumber(rv.Int()))
	case reflect.Float32:
	case reflect.Float64:
		table.Append(lua.LNumber(rv.Float()))
	case reflect.String:
		table.Append(lua.LString(rv.String()))
	case reflect.Slice:
		innertable := L.NewTable()
		for i := 0; i < rv.Len(); i++ {
			appendsomething(L, innertable, rv.Index(i))
		}
		table.Append(innertable)
	default:
		ud := L.NewUserData()
		ud.Value = rv.Interface()
		table.Append(ud)
	}
}
